#/* Copyright 2008-2012 Research Foundation State University of New York   */

#/* This file is part of QUB Express.                                      */

#/* QUB Express is free software; you can redistribute it and/or modify    */
#/* it under the terms of the GNU General Public License as published by   */
#/* the Free Software Foundation, either version 3 of the License, or      */
#/* (at your option) any later version.                                    */

#/* QUB Express is distributed in the hope that it will be useful,         */
#/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
#/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
#/* GNU General Public License for more details.                           */

#/* You should have received a copy of the GNU General Public License,     */
#/* named LICENSE.txt, in the QUB Express program directory.  If not, see  */
#/* <http://www.gnu.org/licenses/>.                                        */

from __future__ import with_statement

from cStringIO import StringIO
import qubx.pyenv
import numpy
import re
from qubx.hist import *
from math import *

# necessary for py2exe/scipy
import scipy.misc
scipy.factorial = scipy.misc.factorial
import scipy.signal

def acceptPow2(x):
    return int(2**max(2, round(log(int(x)) / log(2))))

def acceptObjectName(x):
    eval(x, qubx.pyenv.env.globals)
    return x

npb_rules = [(re.compile(r"hann", re.I), 1.5),
             (re.compile(r"hamm", re.I), 1.36),
             (re.compile(r"blackman", re.I), 1.7),
             (re.compile(r"flattop", re.I), 3.77)]
def guess_noise_power_bandwidth(expr):
    for regexp, npb in npb_rules:
        if regexp.search(expr):
            return npb
    return 1.0 # e.g. None, uniform
npb = memoize(guess_noise_power_bandwidth)


@Propertied(Property('fft_bins', 256, 'number of FFT bins'),
            Property('fft_window', 'numpy.hanning', 'func(N) -> array(shape=(N,)) e.g. numpy.hanning'),
            Property('noise_power_bandwidth', 1.5, 'depends on the window function; common ones are autodetected')
            )
class SpectrumFace(DataHistFace):
    __explore_featured = ['controls', 'copy_to_clipboard', 'copy_image_to_clipboard', 'save_to']
    def __init__(self, name='Spectrum', global_name=""):
        super(SpectrumFace, self).__init__(name, global_name)
        self.__ref = Reffer()
        self.propertied_connect_settings('Spectrum')
        self.plot_count = 1
        self.path = None

        self.controls = qubx.faces.Face("Spectrum", global_name='QubX.About.Spectrum')
        self.controls.show()
        line = pack_item(gtk.HBox(True), self.controls)
        pack_label('Bins:', line)
        self.txtBins = pack_item(qubx.GTK.NumEntry(self.fft_bins, acceptPow2), line)
        self.propertied_connect_NumEntry('fft_bins', self.txtBins)
        pack_label('Window:', self.controls)
        self.mnuWindow = pack_item(qubx.GTK.SuggestiveComboBox(self.fft_window, acceptObjectName), self.controls)
        self.mnuWindow.OnChange += self.__ref(self.__onChangeWindow)
        self.mnuWindow.OnPopup += self.__ref(self.__onPopupWindow)
        self.mnuWindow.OnSuggest += self.__ref(self.__onSuggestWindow)
        pack_label('Noise power bandwith:', self.controls).set_tooltip_text("Auto-detected for common window names")
        self.txtNPB = pack_item(qubx.GTK.NumEntry(self.noise_power_bandwidth, acceptFloatGreaterThan(0.0)), self.controls)
        self.propertied_connect_NumEntry('noise_power_bandwidth', self.txtNPB)
        line = pack_item(gtk.HBox(True), self.controls)
        self.btnSave = pack_button('Save...', line, on_click=self.__onClickSave)
        self.btnCopy = pack_button('Copy', line, on_click=self.__onClickCopy)
        self.btnCopyImage = pack_button('Image...', line, on_click=self.__onClickCopyImage)
    def propertied_set(self, value, name):
        super(SpectrumFace, self).propertied_set(value, name)
        if name == 'fft_window':
            self.mnuWindow.value = value
            self.noise_power_bandwidth = npb(value)
        self.update()
    def __onChangeWindow(self, txt, val):
        self.propertied_on_user_set('fft_window', val)
    def __onPopupWindow(self, txt, suggest):
        suggest("numpy.hanning")
        suggest("numpy.hamming")
        suggest("numpy.blackman")
        suggest("scipy.signal.flattop")
        suggest("None")
    def __onSuggestWindow(self, txt, val):
        self.fft_window = val
    def __onClickSave(self, btn):
        dlg = gtk.FileChooserDialog('Save as...', None, gtk.FILE_CHOOSER_ACTION_SAVE,
                                    buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,gtk.STOCK_OPEN,gtk.RESPONSE_OK))
        dlg.set_default_response(gtk.RESPONSE_OK)
        filter = gtk.FileFilter()
        filter.set_name("text files")
        filter.add_pattern("*.txt")
        dlg.add_filter(filter)
        if self.path: dlg.set_current_folder(self.path)
        dlg.set_filename('Spectrum ')
        response = dlg.run()
        if response == gtk.RESPONSE_OK:
            try:
                fname = add_ext_if_none(dlg.get_filename(), '.txt')
                if self.global_name:
                    qubx.pyenv.env.scriptable_if_matching('%s.save_to(open(%s, "w"))' % (self.global_name, repr(fname)),
                                                          [(self.global_name, self)])
                self.save_to(open(fname, 'w'))
            except Exception, e:
                print "Saving %s: %s" % (fname, e)
            self.path = dlg.get_current_folder()
        dlg.destroy()
    def __onClickCopy(self, btn):
        if self.global_name:
            qubx.pyenv.env.scriptable_if_matching('%s.copy_to_clipboard()' % self.global_name,
                                                  [(self.global_name, self)])
        self.copy_to_clipboard()
    def copy_to_clipboard(self):
        buf = cStringIO.StringIO()
        self.save_to(buf)
        gtk.clipboard_get('CLIPBOARD').set_text(buf.getvalue())
    def __onClickCopyImage(self, btn):
        if gtk.RESPONSE_ACCEPT == TheSpectrumCopyDialog.run(self.plots[0]):
            w = TheSpectrumCopyDialog.copy_width
            h = TheSpectrumCopyDialog.copy_height
            if self.global_name:
                qubx.pyenv.env.scriptable_if_matching('%s.copy_image_to_clipboard(%s, %s)' %
                                                      (self.global_name, repr(w), repr(h)),
                                                      [(self.global_name, self)])
            self.copy_image_to_clipboard(w, h)
    def copy_image_to_clipboard(self, w, h):
        pixmap = gdk.Pixmap(self.window, w, h, -1)
        ctx = pixmap.cairo_create()
        ctx.set_source_rgb(1,1,1)
        ctx.paint()
        self.plots[0].draw_to_context(ctx, w, h)
        del ctx
        pixbuf = gdk.Pixbuf(gdk.COLORSPACE_RGB, False, 8, w, h)
        pixbuf.get_from_drawable(pixmap, gdk.colormap_get_system(), 0, 0, 0, 0, -1, -1)
        clipboard = gtk.clipboard_get('CLIPBOARD')
        clipboard.set_image(pixbuf)
        del pixmap
        del pixbuf
        qubx.pyenv.env.gc_collect_on_idle()
        self.plots[0].redraw_canvas() # re-calculate geometry for mouse
    def onShow(self, showing):
        DataHistFace.onShow(self, showing)
        if showing:
            self.QubX.About.append_face(self.controls)
            self.QubX.About.show_face('Spectrum')
        else:
            self.QubX.About.remove_face( self.QubX.About.index('Spectrum') )
    def save_to(self, f):
        """Writes the histogram as tab-separated text to a.file-like object"""
        f.write(self.plots[0].nbChart.to_txt())
    def on_change_data(self):
        """Determines if update() is needed"""
        if self.QubX.Data.view:
            self.update()
            # self.robot.interrupt()
    def robot_update(self):
        """Calculates a new spectrum."""
        with self.robot.main_hold:
            segs = self.dataSource.get_segmentation()
            nbin = self.fft_bins
            window_f = eval(self.fft_window, qubx.pyenv.env.globals)
            window = window_f(nbin).astype('float32') if window_f else numpy.ones(dtype='float32', shape=(nbin,))
            npb = self.noise_power_bandwidth
            n = sum(seg.n for seg in segs)
        if n == 0: return
        units = segs[0].file.signals[segs[0].signal, 'Units']
        bins = numpy.fft.fftfreq(nbin, segs[0].sampling)[:nbin/2] # discard negative freqs
        
        finished = 0
        spec = MovingWindowSpectrum(window, npb, overlap=2, delta_f=bins[1])
        for seg in segs:
            for frame in generate_sub_samples(nbin/2, self.dataSource.gen_samples(seg.chunks, self.robot.main_hold)):
                if frame is None: # excluded region -- treating it like segment break
                    spec.flush()
                else:
                    spec.add_frame(frame)
                    finished += len(frame)
                    self.robot.progress = finished * 99.0 / n
            spec.flush()
        bars = spec.finish_average()
        self.robot.progress = 100.0

        with self.robot.main_hold:
            self.histrec[0][:] = [HistRec('Power Spectrum Distribution',
                                          'Spectrum', bins, bars, nb_headers=['Hz', '%s^2 rms / Hz' % units],
                                          nb_xlabel='Hz', nb_ylabel='PSD')]
        gobject.idle_add(self.update_plots)

def generate_sub_samples(sub_size, chunks):
    # DataSource.gen_samples has to pause the main thread each time,
    # so it's faster to cache a lot and subdivide it here in-thread
    for chunk in chunks:
        if chunk.included:
            samples = chunk.samples
            n = chunk.n
            i = 0
            while i < n:
                k = min(sub_size, n-i)
                yield samples[i:i+k]
                i += k
        else:
            yield None

class MovingWindowSpectrum(object):
    def __init__(self, window, noise_power_bandwidth, overlap, delta_f):
        self.window = window
        self.npb_correction = 1.0 / (noise_power_bandwidth * delta_f)
        self.overlap = overlap
        self.nbin = len(window)
        self.frame = zeros(shape=(len(window)/2,), dtype='float64')
        self.total_occupancy = 0.0 # like frame count, except first frame is only (1/overlap) full, so contributes that fraction to total_occupancy
        self.data_window = zeros(shape=window.shape, dtype='float32')
        self.piece_size = self.nbin / overlap
        self.data_insert = self.nbin - self.piece_size
        self.data_present = self.nbin
    def add_frame(self, frame, occupied=True):
        n_kept = min(self.data_insert, self.nbin - self.data_present)
        n_added = len(frame)
        if n_kept:
            self.data_window[self.data_insert-n_kept:self.data_insert] = self.data_window[self.nbin-n_kept:self.nbin]
        if n_added:
            self.data_window[self.data_insert:self.data_insert+n_added] = frame
        if n_added < self.piece_size:
            self.data_window[self.nbin - (self.piece_size - n_added):] = 0
        self.data_present = self.data_insert - n_kept
        occupancy = (n_kept + n_added) * 1.0 / self.nbin
        self.total_occupancy += occupancy
        if occupancy:
            psd = numpy.abs(numpy.fft.rfft(self.window * self.data_window))**2
            psd *= occupancy * self.npb_correction / self.nbin
            self.frame += psd[:len(self.frame)]
    def flush(self):
        for i in xrange(self.overlap-1):
            self.add_frame([])
    def finish_average(self):
        if self.total_occupancy:
            self.frame /= self.total_occupancy
            self.total_occupancy = None
        return self.frame


class SpectrumCopyDialog(qubx.GTK.CopyDialog):
    def __init__(self, parent=None):
        qubx.GTK.CopyDialog.__init__(self, 'Copy histogram image', parent)
    def run(self, view):
        response = qubx.GTK.CopyDialog.run(self, view)
        return response

TheSpectrumCopyDialog = SpectrumCopyDialog()



